/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */

package fr.amap.lidar.amapvox.visualization.views;

import fr.amap.commons.util.swing.jprogress.JProgressEvent;
import fr.amap.commons.util.swing.jprogress.JProgressFrame;
import java.awt.Color;
import java.io.IOException;
import java.io.InputStream;
import org.joml.Vector4f;
import fr.amap.lidar.amapvox.visualization.opengl.entities.base.Entity;
import fr.amap.lidar.amapvox.visualization.opengl.entities.base.Scene;
import fr.amap.lidar.amapvox.visualization.opengl.entities.extended.Camera;
import fr.amap.lidar.amapvox.visualization.opengl.entities.extended.CameraController;
import fr.amap.lidar.amapvox.visualization.opengl.entities.extended.MeshRenderer;
import fr.amap.lidar.amapvox.visualization.opengl.entities.extended.VoxelSpace;
import fr.amap.lidar.amapvox.visualization.opengl.entities.util.VoxelSpaceFile;
import fr.amap.lidar.amapvox.visualization.opengl.resources.materials.StandardMaterial;
import fr.amap.lidar.amapvox.visualization.opengl.resources.meshes.DrawingType;
import fr.amap.lidar.amapvox.visualization.opengl.resources.meshes.Geometry;
import fr.amap.lidar.amapvox.visualization.opengl.resources.meshes.Mesh;
import fr.amap.lidar.amapvox.visualization.opengl.resources.postprocesses.extended.GammaCorrectionPostProcess;
import fr.amap.lidar.amapvox.visualization.opengl.resources.postprocesses.extended.SharpenPostProcess;
import fr.amap.lidar.amapvox.visualization.opengl.resources.textures.Texture2D;
import fr.amap.lidar.amapvox.visualization.opengl.resources.textures.TextureFormat;
import fr.amap.lidar.amapvox.visualization.opengl.util.ColorBars;
import fr.amap.lidar.amapvox.visualization.opengl.util.Skyboxes;
import fr.amap.lidar.amapvox.visualization.util.ProjectProperties;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Consumer;

/**
 * 3D scene of the visualization project.
 */
public class VisualizationScene extends Scene {
    
    /* Resources */
    private Mesh cubeMesh;
    private Mesh boundingboxMesh;
    private Mesh axisMesh;
    private Mesh gridMesh;
    private Texture2D axisAlbedo;
    private Texture2D voxelAlbedo;
    
    /* Entities */
    private Entity cameraControllerEntity;
    private CameraController cameraController;
    
    private Entity mainCameraEntity;
    private Camera mainCamera;
    
    private Entity miniCameraEntity;
    private Camera miniCamera;
    
    private Entity axisEntity;
    private MeshRenderer axisRenderer;
    
    private Entity gridEntity;
    private MeshRenderer gridRenderer;
    
    private Entity boundingboxEntity;
    private MeshRenderer boundingboxRenderer;
    
    private Entity voxelSpaceEntity;
    private VoxelSpace voxelSpace;
    
    private List<Consumer<VoxelSpaceFile>> changeFileCallback;
    
    public VisualizationScene() throws IOException {
        super();
        initResources();
        initEntities();
        this.changeFileCallback = new ArrayList();
    }
    
    public final void setHideGrid(final boolean hiden) {
        gridRenderer.getMaterial().setHiden(hiden);
    }
    
    public final void setHideBoundingBox(final boolean hiden) {
        boundingboxRenderer.getMaterial().setHiden(hiden);
    }
    
    public final void setVoxelSpaceFile(final InputStream in, final String filepath,
            final JProgressFrame progressFrame) {
        if (voxelSpace != null) {
            final VoxelSpaceFile file = new VoxelSpaceFile();
            file.setPath(filepath);
            
            if (progressFrame != null) {
                file.addProgressListener((JProgressEvent e) -> {
                    progressFrame.setProgression(e.getValue(), e.getMax(), e.getMessage());
                });
            }
                        
            file.parse(in);
            voxelSpace.setFile(file);
            mainCameraEntity.getTransform().getPosition().z = 
                voxelSpace.getCameraRetreat(60);
            somethingsModified = true;
            
            gridEntity.getTransform().getPosition().z = 1;
            gridEntity.getTransform().getPosition().y = voxelSpace.getGroundHeightPosition();
            
            boundingboxEntity.getTransform().getScale().set(file.getWidth(), file.getDepth(), file.getHeight());
                        
            for (Consumer<VoxelSpaceFile> consumer : changeFileCallback) {
                consumer.accept(file);
            }
        }
    }

    public void setChangeFileCallback(final Consumer<VoxelSpaceFile> changeFileCallback) {
        this.changeFileCallback.add(changeFileCallback);
    }
    
    public final VoxelSpace getVoxelSpaceComponent() {
        return voxelSpace;
    }
    
    public final CameraController getCameraController() {
        return cameraController;
    }
    
    private void initResources() throws IOException {
        cubeMesh = new Mesh("CubeMesh");
        cubeMesh.setGeometry(Geometry.loadFromOBJ("/meshes/cube.obj"));
        
        boundingboxMesh = new Mesh("BoundingboxMesh");
        boundingboxMesh.setGeometry(Geometry.loadFromOBJ("/meshes/boundingbox.obj"));
        boundingboxMesh.setDrawingType(DrawingType.LINES);
        
        gridMesh = new Mesh("GridMesh");
        gridMesh.setGeometry(Geometry.loadFromOBJ("/meshes/grid.obj"));
        gridMesh.setDrawingType(DrawingType.LINES);

        axisMesh = new Mesh("AxisMesh");
        axisMesh.setGeometry(Geometry.loadFromOBJ("/meshes/axis.obj"));

        axisAlbedo = new Texture2D("AxisAlbedo", 256, 256,
                TextureFormat.RGBA, true, "/textures/axisAlbedo.png");

        voxelAlbedo = new Texture2D("VoxelAlbedo", 256, 256,
                TextureFormat.RGBA, true, "/textures/voxelAlbedo.png");

        setBackground(Skyboxes.getSky(ProjectProperties.getProperty("background", "Simple")));
    }
    
    private void initEntities() throws IOException {
        /* Create the voxel space */
        voxelSpaceEntity = new Entity();
        voxelSpaceEntity.setLayout("voxel");
        voxelSpace = new VoxelSpace(cubeMesh);
        voxelSpace.setAlbedo(voxelAlbedo);
        voxelSpace.setColorBar(ColorBars.getColorBar(
                ProjectProperties.getProperty("colorBar", "Equatorial")));
        voxelSpace.setDrawFastEnabled(Boolean.valueOf(
                ProjectProperties.getProperty("drawFastEnabled", "true")));
        voxelSpaceEntity.addComponent(voxelSpace);
        addEntity(voxelSpaceEntity);
        
        /* Create the camera controller */
        cameraControllerEntity = new Entity();
        cameraController = new CameraController();
        cameraControllerEntity.addComponent(cameraController);
        addEntity(cameraControllerEntity);

        /* Create the main camera */
        mainCameraEntity = new Entity();
        mainCameraEntity.setParent(cameraControllerEntity);
        
        mainCamera = new Camera();
        mainCamera.setZoomEnabled(true);
        mainCamera.addLayout("background");
        mainCamera.addLayout("voxel");
        mainCamera.addPostProcess(0, new SharpenPostProcess());
//        mainCamera.addPostProcess(1, new TonemappingPostProcess());
        //mainCamera.addPostProcess(2, new FXAAPostProcess());
        mainCamera.addPostProcess(3, new GammaCorrectionPostProcess());
        mainCamera.setOnZoomCallback((fov) -> {
            voxelSpace.setCameraFOV(fov);
        });
        
        mainCameraEntity.addComponent(mainCamera);
        addEntity(mainCameraEntity);
        
        /* Create the axis */
        axisRenderer = new MeshRenderer();
        final StandardMaterial axisMaterial = new StandardMaterial();
        axisMaterial.setAlbedo(axisAlbedo);
        axisMaterial.setMainColor(Color.WHITE);
        axisRenderer.setMesh(axisMesh);
        axisRenderer.setMaterial(axisMaterial);
        
        axisEntity = new Entity();
        axisEntity.setLayout("axis");
        axisEntity.addComponent(axisRenderer);
        addEntity(axisEntity);
        
        /* Create the grid */
        gridRenderer = new MeshRenderer();
        final StandardMaterial gridMaterial = new StandardMaterial();
        gridMaterial.setMainColor(Color.GRAY);
        gridRenderer.setMaterial(gridMaterial);
        gridRenderer.setMesh(gridMesh);
        
        gridEntity = new Entity();
        gridEntity.setLayout("voxel");
        gridEntity.addComponent(gridRenderer);
        addEntity(gridEntity);
        
        /* Create the bounding box */
        boundingboxRenderer = new MeshRenderer();
        final StandardMaterial boundingboxMaterial = new StandardMaterial();
        boundingboxMaterial.setMainColor(Color.GRAY);
        boundingboxRenderer.setMaterial(boundingboxMaterial);
        boundingboxRenderer.setMesh(boundingboxMesh);
        
        boundingboxEntity = new Entity();
        boundingboxEntity.addComponent(boundingboxRenderer);
        boundingboxEntity.setLayout("voxel");
        addEntity(boundingboxEntity);
        
        /* Create the mini camera */
        miniCamera = new Camera();
        miniCamera.addLayout("axis");
        miniCamera.setProportions(new Vector4f(0.79f, 0.01f, 0.2f, 0.2f));
        //miniCamera.addPostProcess(0, new SharpenPostProcess());
        //miniCamera.addPostProcess(3, new GammaCorrectionPostProcess());
        
        miniCameraEntity = new Entity();
        miniCameraEntity.setParent(cameraControllerEntity);
        miniCameraEntity.getTransform().getPosition().z = 2.5f;
        miniCameraEntity.addComponent(miniCamera);
        addEntity(miniCameraEntity);
    }

}
